
; MSX ROM Loader for Musical Memory Mapper

; Version 0.81 by GDX

; Assembled with zasm cross assembler

; http://sourceforge.net/projects/zasm/

; *S 14	; Remove the semicolon at the start line when using GEN80 assembler

LF:	equ	0Ah
CR:	equ	0Dh
BDOS:	equ	00005h
CALSLT:	equ	0001Ch
DMA:	equ	00080h
ENASLT:	equ	00024h
FCB:	equ	0005ch
EXPTBL:	equ	0FCC1h
MSXVER:	equ	0002dh
RAMAD2:	equ	0F343h
RDSLT:	equ	0000Ch
SUBROM:	equ	0015Ch
WRRES:	equ	0017Dh
WRSLT:	equ	00014h


	org	0100h

	jp	Prog
MainMessage:

	db	"ROM Loader for MMM",LF,CR
	db	"v.0.81 by GDX",LF,CR,LF,CR,"$"
	db	"Loading...",LF,CR,"$"

FileNotFound:
	db	"File Not Found!",LF,CR,"$"
BadFile:
	db	"Bad file!",LF,CR,"$"
No_MMM:
	db	"MMM Not Found!",LF,CR,"$"
Mapped_ROM:
	db	"Mapped Roms are not supported!",LF,CR,"$"
CHK_RAM:
	db	0
FileSize:
	db	0	; File Size value
KBuf:
	db	0	; Tempory value
MMM_Slot:
	db	0	; Store MMM Slot value
Page0:
	db	0	; MMM page value for bank 0000h~3fffh
Page1:
	db	0	; MMM page value for bank 4000h~7fffh
Page2:
	db	0	; MMM page value for bank 8000h~ffffh
Page3:
	db	0	; MMM page value for bank C000h~ffffh
RAMAD3_New:
	db	0	; Slot of internal RAM on bank c000h~ffffh
MSXVER_NEW:
	db	0	; MSX version
EXPTBL_New:
	db	0,0,0,0	; Slot primary 0, 1, 2 and 3 flags


; Musical Memory Mapper detection

MMM_Srch:
	ld	l,0FFh
	ld	b,16

MMM_Srch_Loop:
	push	hl

	ld	a,b
	dec	a
	and	3

	ld	hl,EXPTBL
	add	a,l
	ld	l,a

	ld	a,b
	dec	a
	or	(hl)

	pop    hl

	jp	m,Sec_SLT	; Jump if secondary Slot

	and	%00001111
	cp	4
	jr	nc,NextMMM_Srch ; Jump if 3-3 Slot

Sec_SLT:
	ld	c,a

	push	bc
	push	hl
	call	ENASLT		; Slot to search
	pop	hl
	pop	bc

	di

	ld	a,080h
	out	(03Ch),a	; Enable acces to registers by memory address mode

	ld	a,(hl)
	and	%00011111
	inc	a

	out	(0FFh),a

	or	%11000000
	cp	(hl)		; Same value?

	ld	a,0
	out	(03Ch),a
	out	(0FFh),a	; Restore initial mapper page on bank 0C000H~0FFFFH 

	ei

	ld	a,c		; A = Found Slot Number (F000SSPP)
	jr	z,MMM_Found

NextMMM_Srch:
	djnz	MMM_Srch_Loop	; Jump if MMM is not found

	ld	a,0FFh		; MMM not found value
	ret

MMM_Found:
	dec	b
	ret

Prog:
	ld	hl,EXPTBL
	ld	de,EXPTBL_New
	ld	bc,4
	ldir			; Copy EXPTBL (Extended Slot flags)

	ld	hl,MSXVER	; MSX version
	ld	a,(EXPTBL)	; Main-ROM slot
	call	RDSLT
	ld	(MSXVER_NEW),a	; MSXVER value

	ld	c,9
	ld	de,MainMessage
	call	BDOS		; Print Main Message

	ld	bc,24		; Prepare the FCB
	ld	de,FCB+13
	ld	hl,FCB+12
	ld	(hl),b
	ldir			; Initialize the second half with zero

	ld	H,080h
	call	MMM_Srch	; MMM detection
;----
;	bit	7,a
;	jr	nz,MMMBit7	; Jump if the slot is extended
;	and	003h		; Mask unused mask
;MMMBit7:
;----	

	ld	(MMM_Slot),a	; Store MMM slot value
	cp	0ffh
	ld	de,No_MMM
	jp	z,Exit		; Jump if MMM not Found

	ld	h,080h
	ld	a,(MMM_Slot)
	call	ENASLT		; Select MMM RAM at Bank 8000h~BFFFh

; ROM Loading

	ld	c,0Fh
	ld	de,FCB
	call	BDOS		; Open file
	ld	hl,1
	ld	(FCB+14),hl	; Record size = 1 byte
	ld	de,FileNotFound
	or	a
	jp	nz,Exit		; Jump if File not found

	ld	c,1ah
	ld	de,04000h
	call	BDOS		; Set disk transfer address (buffer start at 04000h)

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

;	push	hl
	ld	hl,FCB+16
	xor	a
	or	(hl)
	inc	hl
	or	(hl)
	inc	hl
	or	(hl)
	inc	hl
	or	(hl)
	ld	(FileSize),a	; Get a simple value to test file size 

	cp	020h
	jp	nz,No_8KB	; Jump if file size <> 02000h
	
	ld	hl,04000h
	ld	de,06000h
	ld	bc,02000h
	ldir			; Make mirror

	jr	Rom8k16k

No_8KB:
	ld	a,(FileSize)
	cp	040h
	jp	nz,No_16KB

	di
Rom8k16k:
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,4
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page2),a	; Store MMM page number for bank 8000h~BFFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM


	ld	a,(08003h)	; Start Rom address
	rla
	rla
	jr	c,Mirror_P1

	ld	a,(08003h)	; Start Rom address
	rla
	jr	c,No_mirror_P1	; Jump if Start Rom adress = 0000h

	ld	a,(08009h)	; Read Most Significant byte of start Basic pointer
	rla
	jr	c,No_mirror_P1

Mirror_P1:
	ld	a,4
	jr	P1
No_mirror_P1:
	ld	a,5
P1:
	ld	(Page1),a	; Store MMM page number for bank 4000h~7FFFh

	ld	a,(08003h)	; Start Rom address
	cp	40h
	ld	a,4
	jr	c,Mirror_P0
	ld	a,5
Mirror_P0:
	ld	(Page0),a	; Store MMM page number for bank 0000h~3FFFh

	ld	a,4
	ld	(Page3),a	; Store MMM page number for bank C000h~FFFFh
	
	jp	MoveRomExec

No_16KB:
	ld	a,(FileSize)
	cp	080h
	jp	nz,No_32KB

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,4
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page1),a	; Store MMM page number for bank 4000h~7FFFh
	ld	(Page3),a	; Store MMM page number for bank C000h~FFFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM
	ei

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,5
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page2),a	; Store MMM page number for bank 8000h~BFFFh
	ld	(Page0),a	; Store MMM page number for bank 0000h~3FFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM

	jp	MoveRomExec

No_32KB:
	ld	a,(FileSize)
	cp	0C0h
	jp	nz,No_48KB

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,4
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page0),a	; Store MMM page number for bank 0000h~3FFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM
	ei

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,5
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page1),a	; Store MMM page number for bank 0000h~3FFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM
	ei

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,6
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page2),a	; Store MMM page number for bank 0000h~3FFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM

	ld	a,7
	ld	(Page3),a	; Store MMM page number for bank C000h~FFFFh

	jp	MoveRomExec

No_48KB:
	ld	a,(FileSize)
	cp	001h
	jp	nz,No_64KB

	ld	a,(04000h)
	cp	041h		; Compare 41h with read value
	jr	nz,No_Mapper	; Jump if No mapped Rom
	ld	a,(04001h)
	cp	042h		; Compare 42h with read value
	jr	nz,No_Mapper	; Jump if No mapped Rom

	ld	de,Mapped_ROM
	jp	Exit		; Exit if the Rom is mapped (Rom header found)

No_Mapper:
	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,4
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page0),a	; Store MMM page number for bank 0000h~3FFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM
	ei

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,5
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page1),a	; Store MMM page number for bank 4000h~7FFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM
	ei

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,6
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page2),a	; Store MMM page number for bank 8000h~BFFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM
	ei

	ld	c,27h
	ld	de,FCB
	ld	hl,04000h	; Number of records to read
	call	BDOS		; Read a block in file

	di
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,7
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	(Page3),a	; Store MMM page number for bank C000h~FFFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,04000h
	ld	de,08000h
	ld	bc,04000h
	ldir			; Move Rom bank to the page of MMM

	jp	MoveRomExec

No_64KB:
	ld	de,BadFile
	jp	Exit		; Jump if bad file

MoveRomExec:
	ld	a,080h
	out	(03ch),a	; Enable acces register via address
	ld	a,2
	ld	(080FEh),a	; Select MMM page 2 on bank 8000h~BFFFh
	ld	a,000h
	out	(03ch),a	; Disable acces register via address	

	ld	hl,PreRom_Exec
	ld	de,08000h
	ld	bc,Rom_Exec-PreRom_Exec
	ldir

	jp	08000h		; Jump to PreRom_Exec
	
PreRom_Exec:
	ld	a,(MSXVER_NEW)
	cp	3
	jr	z,MSX_tR	; Jump if MSX2+ or newer (Keep Internal RAM if MSX Turbo R)

; ---
; Search and select internal RAM on bank C000h~FFFFh

	ld	b,15
Srch_RAM_Loop:
	ld	a,b
	rra
	rra
	rra
	and	0c0h		; Keep primary slot number
	ld	c,a
	in	a,(0a8h)
	and	03fh		; Reset primary slot number of bank C000h~FFFFh
	or	c
	out	(0a8h),a	; Select primary slot on bank C000h~FFFFh

	ld	hl,EXPTBL_New
	ld	d,0
	ld	a,b
	and	03h		; Reverse primary slot number bits
	ld	e,a
	add	hl,de

	bit	7,(hl)
	jr	z,No_bit7	; Jump if the slot is not extended

	ld	a,b
	rla
	rla
	rla
	rla
	and	0c0h		; Keep secondary slot number
	ld	c,a
	ld	a,(0ffffh)
	cpl
	and	03fh
	or	c
	ld	(0ffffh),a	; Select secondary slot on Bank C000h~FFFFh

No_bit7:
	ld	a,(0f91eh)
	ld	(KBuf),a	; Save value at 0f91eh
	ld	a,41h
	ld	(0f91eh),a	; Write 41h
	ld	a,(0f91eh)
	cp	041h		; Compare 41h with read value
	jr	nz,noRam
	inc	a
	ld	(0f91eh),a	; Write 42h
	ld	a,(0f91eh)
	cp	042h		; Compare 42h with read value
	
	ld	a,(KBuf)
	ld	(0f91eh),a	; Restore value at 0f91eh

	jr	z,cmp_MMM_Slt	; Jump if RAM
noRam:
	djnz	Srch_RAM_Loop	; Next slot

cmp_MMM_Slt:
	ld	a,(MMM_Slot)
	ld	c,a
	
	ld	a,b
	or	(HL)		; Set bit 7 if Secondary Slot
	bit	7,a
	jr	nz,Bit7		; Jump if the slot is extended
	and	003h		; Mask unused mask
Bit7:
	cp	c		; Compare avec le Slot de MMM
	jr	z,noRam		; Jump to djnz of Srch_RAM_Loop if it is the MMM Slot

	ld	(RAMAD3_New),a	; Save RAM slot number
; ---
MSX_tR:	

	ld	hl,MMM_Slot
	ld	de,0C000h
	ld	bc,6
	ldir			; Move page & slot numbers to c000h

	ld	hl,Rom_Exec
	ld	de,0c006h
	ld	bc,Exit-Rom_Exec
	ldir

	ld	a,(EXPTBL_New)
	ld	c,a		; C = Main-Rom Slot number
	in	a,(0a8h)
	ld	b,a		; Store actual Primary Slots register
	and	03Ch
	out	(0a8h),a	; Select primary Slot 0 on 0~3FFFh and 0c000h~FFFFh banks
	ld	a,c
	and	080h
	jr	z,No_Sec_slt	; Jump if primary slot
	ld	a,(0FFFFh)
	cpl
	and	0FCh
	ld	(0FFFFh),a	; Select secondary Slot 0 on 0~3FFFh (Main-ROM)
No_Sec_slt:
	ld	a,b		; Restore primary slots register
	and	0fch
	out	(0a8h),a	; Restore before Primary Slot 3 to register
	
	jp	0C006h

Rom_Exec:
	ld	a,080h
	out	(03ch),a	; Enable acces register via address

	ld	a,(0C001h)	; restore 0FCh page value to A
	ld	(80FCh),a	; Select MMM page on bank 0000h~3FFFh
	ld	a,(0C002h)	; restore 0FDh page value to A
	ld	(80FDh),a	; Select MMM page on bank 4000h~7FFFh
	ld	a,(0C003h)	; restore 0FEh page value to A
	ld	(080FEh),a	; Select MMM page on bank 8000h~BFFFh
	ld	a,(0C004h)	; restore 0FFh page value to A
	ld	(80FFh),a	; Select MMM page on bank C000h~FFFFh

	ld	a,06Fh
	ld	(803Ch),a	; Write protect the MMM banks and disable access via I/O
	ld	a,000h
	out	(03ch),a

	ld	b,4
	ld	a,09Fh
Volume0:
	out	(03Fh),a	; Set volume of
	add	a,020h		; SN76489AN
	djnz	Volume0	; to zero

	rst	0		; Execute the ROM by a soft reset 
	
Exit:
	push	de

	ld	h,080h
	ld	a,(RAMAD2)
	call	ENASLT		; Select MMM RAM at Bank 8000h~BFFFh

	pop	de
	ld	c,9
	call	BDOS		; Print Message

	rst	0		; Back to MSX-DOS


	